"""
Entidades Event Sourcing
"""
import datetime
from clitellum_evs.rejection.engines import RejectionEventException, IgnoreRejectionEngine

EPOCH = datetime.datetime.utcfromtimestamp(0)

class DomainEvent(object):
    """
    Interface de Evento de dominio
    """

    def __init__(self, name, id, data):
        self._name = name
        self._id = id
        self._data = data
        self._version = (datetime.datetime.utcnow() - EPOCH).total_seconds() * 1000000

    def get_name(self):
        """
        Devuelve el nombre de la transicion asociada al evento
        """
        return self._name

    def get_id(self):
        return self._id

    def to_json(self):
        return self._data

    def apply(self, entity):
        """
        Realiza la aplicacion del evento sobre la entidad
        """
        pass

    def set_version(self, version):
        self._version = version

    def get_version(self):
        return self._version

    def can_apply(self, entity):
        """
        Indica si se puede aplicar el envento sobre la entidad
        """
        return True

class AggregateRoot(object):
    """
    Clase aggregate root, para la aplicacion de los eventos
    """
    def __init__(self, id, entity, state_machine, initial_state, evs_repository, snapshot_builder):
        self._state_machine = state_machine
        self._state_machine.set_state(initial_state)
        self._root = entity
        self._applied_events = []
        self._id = id
        self._version = 0
        self._rejection_engine = IgnoreRejectionEngine()
        self._evs_repository = evs_repository
        self._snapshot_builder = snapshot_builder

    def set_rejection_engine(self, engine):
        '''
        Establece el motor de rechazo
        '''
        self._rejection_engine = engine

    def get_rejection_engine(self):
        '''
        Devuelve el motor de rechazo
        '''
        return self._rejection_engine

    def add_event(self, event):
        '''
        Anade y aplica un evento al root
        '''
        self._evs_repository.save_event(event, self._id)
        self.apply_event(event)

    def apply_event(self, event):
        """
        Aplica el evento sobre la raiz
        """
        if event.get_version() < self._version:
            self._rejection_engine.reject(self, event)
            return

        if self._state_machine.transite(event.get_name()):
            if event.can_apply(self._root):
                event.apply(self._root)
                self._version = event.get_version()
                self._applied_events.append(event)
                return

        self._rejection_engine.reject(self, event)

    def apply_events(self, events):
        """
        Aplica una lista de eventos sobre la raiz
        """
        for event in events:
            try:
                self.apply_event(event)

            except RejectionEventException:
                break

    def get_applied_events(self):
        '''
        Devuelve los eventos que se han aplicado
        '''
        return self._applied_events

    def get_state(self):
        """
        Devuelve el estado de la entidad
        """
        return self._state_machine.get_state()

    def get_id(self):
        return self._id

    def get_entity(self):
        return self._root

    def get_version(self):
        return self._version

    def set_version(self, version):
        self._version = version

    def create_snapshot(self):
        '''
        Crea el snahpshot de la entidad
        '''
        self._snapshot_builder.build(self)



import pymongo

class EvsRepository(object):
    '''
    Repositorio de eventos
    '''

    def __init__(self, database, collection_name):
        self._database = database
        self._collection_name = collection_name
        self._entity_col = self._database.get_collection(self._collection_name + '_entity')
        self._applied_col = self._database.get_collection(self._collection_name + '_applied')
        self._rejected_col = self._database.get_collection(self._collection_name + '_rejected')
        self._events_col = self._database.get_collection(self._collection_name + '_events')

    def get_entity(self, query):
        '''
        Devuelve la entidad segun la query especificada
        '''
        return self._entity_col.find_one(query)

    def get_entity_by_id(self, entity_id):
        '''
        Devuelve la entidad por el identificador especificado
        '''
        return self._entity_col.find_one({"_id" : entity_id})


    def get_events(self, entity_id):
        '''
        Devuelve los eventos asociados a la entidad especificada
        '''
        return self._events_col.find({'entity_id': entity_id}) \
            .sort("version", pymongo.ASCENDING)

    def save_event(self, event, entity_id):
        '''
        Guarda el evento de la entidad especificada
        '''
        event_db = self._create_event_db(event, entity_id)
        self._events_col.update({'_id': event.get_id()}, event_db, upsert=True)

    def save_applied_event(self, event, entity_id):
        '''
        Guarda los eventos aplicados de la entidad
        '''
        event_db = self._create_event_db(event, entity_id)
        self._applied_col.update({'_id': event.get_id()}, event_db, upsert=True)
        self._events_col.remove({'_id': event.get_id()})

    def save_rejected_event(self, event, entity_id):
        '''
        Guarda los eventos rechazados de la entidad
        '''
        event_db = self._create_event_db(event, entity_id)
        self._rejected_col.update({'_id': event.get_id()}, event_db, upsert=True)
        self._events_col.remove({'_id': event.get_id()})

    def _create_event_db(self, event, entity_id):
        event_db = {}
        if not event.get_id() is None:
            event_db['_id'] = event.get_id()
        event_db['event'] = event.to_json()
        event_db['entity_id'] = entity_id
        event_db['version'] = event.get_version()
        event_db['_t'] = {"module": event.__module__,
                          "name": event.__class__.__name__}
        return event_db

    def save_entity(self, entity_id, entity, state, version):
        entity_db = {}
        entity_db['snapshot'] = entity
        entity_db['state'] = state
        entity_db['version'] = version
        self._entity_col.update({'_id' : entity_id}, \
                                             entity_db, upsert=True)




'''
Motores de rechazo
'''

class RejectionEventException(Exception):
    '''
    Excepcion que se lanza cuando un evento ha sido rechazado en alguna transición
    '''
    def __init__(self, root, event):
        msg = "Error al procesar el evento %s, en la entidad %s" % (event.get_name(), str(root.get_id()))
        Exception.__init__(self, msg)

class RejectionEngine(object):
    '''
    Clase base de los motores de rechazo
    '''
    def __init__(self):
        self._rejected_events = []

    def get_rejected_events(self):
        '''
        Devuelve la lista de los eventos rechazados
        '''
        return self._rejected_events

    def reject(self, root, event):
        '''
        Rechazo del evento
        '''
        raise NotImplementedError()

    def execute(self):
        '''
        Ejecuta la logica final del motor de rechazo
        '''
        raise NotImplementedError()

class IgnoreRejectionEngine(RejectionEngine):
    '''
    Define el motor de rechazo que no realiza ninguna accion cuando se rechaza un evento.
    '''
    def __init__(self):
        RejectionEngine.__init__(self)

    def reject(self, root, event):
        self._rejected_events.append(event)

    def execute(self):
        pass

class StopOnErrorRejectionEngine(RejectionEngine):
    '''
    Define el motor de rechazo de paro, cuando llega alguna transición que se ha podido aplicar
    no se siguen aplicando mas eventos.
    '''
    def __init__(self):
        RejectionEngine.__init__(self)

    def reject(self, root, event):
        self._rejected_events.append(event)
        raise RejectionEventException(root, event)

    def execute(self):
        pass



"""
Entidades Builder
"""
import sys
from clitellum_evs.repository import EvsRepository
from clitellum_evs.entities import AggregateRoot
from clitellum_evs.snapshot import SnapshotBuilder
from bson import ObjectId
from clitellum_evs.rejection.engines import IgnoreRejectionEngine, StopOnErrorRejectionEngine

class AggregateRootBuilder(object):
    """
    Aggregate Root Builder for finite state machines.
    """
    def __init__(self, evs_repository, state_machine_class):
        self._state_machine_class = state_machine_class
        self._id = None
        self._state = None
        self._query = None
        self._events = []
        self._rejection_engine = IgnoreRejectionEngine()
        self._evs_repository = evs_repository
        self._snapshot_builder = SnapshotBuilder(evs_repository)

    @classmethod
    def create(cls, database, collection_name, state_machine_class):
        """

        :param database: base de datos.
        :param collection_name: nombre de la coleccion donde reflejar los eventos.
        :param state_machine_class: nombre de la clase tipo de la maquina de estados.
        :return: src.builder.AggregateRootBuilder
        """
        repository = EvsRepository(database, collection_name)
        return AggregateRootBuilder(repository, state_machine_class)

    def set_entity(self, obj_id):
        """
        Setea el id de la entidad.
        :param obj_id: identificador de la entidad
        """
        self._id = obj_id
        return self
    
    def set_entity_from_query(self, query):
        """
        Setea el id de la entidad.
        :param id: identificador de la entidad
        """
        self._query = query
        return self

    def set_initial_state(self, state):
        """
        Setea el estado inicial
        :param state: estado
        """
        self._state = state
        return self

    def set_stop_on_rejection(self):
        '''
        Establece el motor de rechazo StopOnErrorRejection, no aplica mas eventos si uno 
        da error
        '''
        self._rejection_engine = StopOnErrorRejectionEngine()
        return self

    def set_snapshot_builder(self, builder):
        """
        Setea el snapshot_builder
        :param builder: snapshot builder
        """
        self._snapshot_builder = builder
        return self

    def add_event(self, event):
        """
        Anade un evento con version.
        :param event: evento
        """
        self._events.append(event)
        return self

    def build(self):
        """
        Hace el build de un Aggregate Root
        :return:
        """
        self._get_id()

        for event in self._events:
            self._evs_repository.save_event(event, self._id)

        entity = self._evs_repository.get_entity_by_id(self._id)
        events = self._evs_repository.get_events(self._id)
        machine = self._state_machine_class()

        if self._state is None:
            self._state = machine.get_initial_state()

        if entity is None:
            root = AggregateRoot(self._id, {}, machine, self._state, self._evs_repository, self._snapshot_builder)
        else:
            root = AggregateRoot(self._id, entity['snapshot'], machine, entity['state'], self._evs_repository, self._snapshot_builder)
            root.set_version(entity['version'])

        root.set_rejection_engine(self._rejection_engine)

        root_events = []
        for event in events:
            event_class = getattr(sys.modules[event['_t']['module']], event['_t']['name'])
            ev = event_class(event['_id'], event['event'])
            ev.set_version = event['version']
            root_events.append(ev)
        
        root.apply_events(root_events)
        return root

    def _get_id(self):
        if not self._id is None:
            return
        if self._query is None:
            self._id = ObjectId()
        ret = self._evs_repository.get_entity(self._query)
        if ret is None:
            self._id = ObjectId()
        else:
            self._id = ret['_id']





from setuptools import setup

setup(
    name='clitellum_evs',
    version='1.2.6',
    packages=['clitellum_evs', 'clitellum_evs.rejection'],
    package_dir={'clitellum_evs': 'src'},
    url='https://gitlab.com/clitellum/clitellum_evs/wikis/home',
    license='APACHE',
    author='Sergio.Bermudez',
    author_email='sbermudezlozano@gmail.com',
    description='Event Sourcing Library',
    keywords="clitellum event sourcing eventsourcing",
    extras_require={
    },
    install_requires=['pymongo']
)



import sys
from clitellum_evs.repository import EvsRepository
from clitellum_evs.rejection.engines import IgnoreRejectionEngine, StopOnErrorRejectionEngine

class SnapshotBuilder(object):
    """
    Creador de imagenes de una entidad en concreto.
    """
    def __init__(self, evs_repository):
        self._evs_repository = evs_repository
        self._remove_rejected_events = True


    def set_remove_rejected_events(self, remove):
        '''
        Estable si al crear el snapshot hay que eliminar los eventos que han sido rechazados
        por defecto no se eliminan de la lista de eventos.
        '''
        self._remove_rejected_events = remove
        return self

    def build(self, root):
        """
        Build
        """
        for event in root.get_applied_events():
            self._evs_repository.save_applied_event(event, root.get_id())

        if self._remove_rejected_events:
            for event in root.get_rejection_engine().get_rejected_events():
                self._evs_repository.save_rejected_event(event, root.get_id())

        self._evs_repository.save_entity(root.get_id(), root.get_entity(), \
                                        root.get_state(), root.get_version())



'''
Maquina de estados del event sourcing
'''

class EventSourcingMachine(object):
    """
    Clase que define la maquina de estados
    """

    def __init__(self):
        self._states = []
        self._transitions = {}
        self._initial_state = None
        self._current_state = None

    def add_state(self, name, initial_state=False):
        """
        Anade estado a los estados validos de la maquina.
        :param name: nombre del estado
        """
        if name in self._states:
            return

        self._states.append(name)
        if initial_state:
            self._initial_state = name
            self._current_state = name

    def add_transition(self, name, from_state, to_state):
        """
        Anade transicion a partir de un nombre y dos estados.
        :param name: nombre de la transicion
        :param from_state: estado
        :param to_state: estado
        """
        if not from_state in self._states or not to_state in self._states:
            return

        if name in self._transitions:
            self._transitions[name][from_state] = to_state
        else:
            self._transitions[name] = {from_state: to_state}

    def set_state(self, state):
        """
        Setea el estado pasado por parametro como estado actual.
        :param state: estado
        """
        self._current_state = state

    def get_initial_state(self):
        """
        Devuelve el estado inicial de la maquina de estados
        """
        return self._initial_state

    def get_state(self):
        """
        Devuelve el estado actual.
        :return: Estado
        """
        return self._current_state

    def transite(self, name):
        """
        Transiciona una transicion si esta definida.
        :param name: nombre de la transicion
        :return: boolean si ha podido ejecutar la transicion.
        """
        if name in self._transitions:
            if self._current_state in self._transitions[name]:
                self._current_state = self._transitions[name][self._current_state]
                return True
        return False
